import os
import sys
import time
import json
import math
import numpy as np
import matplotlib.pyplot as plt

from tqdm import tqdm
from lxml import etree
from colorama import Fore
from collections import Counter
from pymatgen.io.vasp.outputs import Vasprun
from itertools import combinations_with_replacement

from vachoppy.utils import CosineDistance

try:
    from mpi4py import MPI
    PARALELL = True
except:
    PARALELL = False
    
BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m' # Green color
RED = '\033[91m'   # Red color
RESET = '\033[0m'  # Reset to default color


class FingerPrint:
    def __init__(self,
                 A,
                 B,
                 poscar,
                 Rmax=10,
                 delta=0.08,
                 sigma=0.03,
                 dirac='g'):
        """
        poscar: path of POSCAR file. (in direct format)
        Rmax: threshold radius
               (valid value: 1-30)
        delta: discretization of fingerprint function 
               (valid value: 0.01-0.2)
        sigma: Gaussian broadening of interatomic distances
               (valid value: 0.01-0.1)
        dirac: 's' for square-form dirac function or 
               'g' for Gaussian broaden dirac function.
        """
        self.A = A
        self.B = B
        self.poscar = poscar
        self.Rmax = Rmax
        self.delta = delta
        self.sigma = sigma
        self.dirac = dirac
        self.R = np.arange(0, self.Rmax, self.delta)
        self.fingerprint = np.zeros_like(self.R)
        self.tolerance = 1e-3
        
        # read poscar
        self.lattice = None
        self.atom_species = None
        self.atom_num = None
        self.position = []
        self.read_poscar()
        
        # Search indices of A and B
        self.idx_A = self.search_atom_index(self.A)
        self.idx_B = self.search_atom_index(self.B)
        
        # volume of unit cell
        self.V_unit = np.abs(
            np.dot(np.cross(self.lattice[0], self.lattice[1]), 
                   self.lattice[2]))
        
        # extended coordinaitons of B
        self.B_ext = None
        self.get_extended_coords_B()

        # calculate fingerprint
        self.get_fingerprint()
    
    
    def read_poscar(self):
        with open(self.poscar, 'r') as f:
            lines = [line.strip() for line in f]
        
        scale = float(lines[1])
        self.lattice = [line.split() for line in lines[2:5]]
        self.lattice = np.array(self.lattice, dtype=float) * scale
        
        self.atom_species = lines[5].split()
        self.atom_num = np.array(lines[6].split(), dtype=int)
        
        # check whether poscar is direct type
        if not lines[7][0].lower() == 'd':
            print('only direct type POSCAR is supported.')
            sys.exit(0)
        
        # parse positions
        start = 8
        for atom, num in zip(self.atom_species, self.atom_num):
            coords = [line.split()[:3] for line in lines[start:start+num]]
            coords = np.array(coords, dtype=float)
            self.position.append({'name': atom, 'num': num, 'coords':coords})
            start += num
           
            
    def gaussian_func(self, x):
        return (1/np.sqrt(2*np.pi*self.sigma**2)) *\
               np.exp(-x**2 / (2*self.sigma**2))
    
    
    def square_func(self, x):
        return np.array(np.abs(x) <= self.sigma)/(2*self.sigma)
    
    
    def dirac_func(self, x):
        if self.dirac[0] == 'g':
            return self.gaussian_func(x)
        elif self.dirac[0] == 's':
            return self.square_func(x)
        else:
            raise ValueError(f"{self.dirac} is not defined.")
            
            
    def search_atom_index(self, atom_name):
        for i, atom in enumerate(self.position):
            if atom['name'] == atom_name:
                return i
        raise ValueError(f"Atom {atom_name} not found in POSCAR file.")
    
    
    def get_extended_coords_B(self):
        # supercells within Rmax
        l_lat = np.linalg.norm(self.lattice, axis=1)
        m = np.floor(self.Rmax / l_lat) + 1
        
        # make 3D grid of supercells
        mx, my, mz = [np.arange(-mi, mi+1) for mi in m]
        shifts = np.array(np.meshgrid(mx, my, mz)).T.reshape(-1, 3)

        # save candidated of B atoms
        coords_B = self.position[self.idx_B]['coords']
        self.B_ext = np.vstack([shifts + coord for coord in coords_B])
             
             
    def get_fingerprint(self):
        for coord_A in self.position[self.idx_A]['coords']:
            self.fingerprint += self.get_fingerprint_i(coord_A)
        
        self.fingerprint *= self.V_unit / self.position[self.idx_A]['num']
        self.fingerprint -= 1
        
        
    def get_fingerprint_i(self, coord_A_i):
        # calculate R_ij
        disp = self.B_ext - coord_A_i
        disp_cart = np.dot(disp[:,:], self.lattice)
        R_ij = np.linalg.norm(disp_cart, axis=1)
        
        # When A=B, i=j should be excluded. (else, diverge)
        if self.idx_A == self.idx_B:
            R_ij[R_ij < self.tolerance] = np.inf
            
        # number of B atoms within Rmax
        N_B = np.sum(R_ij <= self.Rmax)
        fingerprint_i = np.zeros_like(self.fingerprint)

        for idx, r in enumerate(self.R):
            dirac_values = self.dirac_func(r - R_ij)
            valid_indices = R_ij <= self.Rmax
            fingerprint_i[idx] = np.sum(
                dirac_values[valid_indices] / R_ij[valid_indices]**2)\
                      / (4 * np.pi * N_B * self.delta)

        return fingerprint_i
    
    
    def plot_fingerprint(self, 
                         disp=True,
                         save=False,
                         label=None,
                         outdir='./',
                         dpi=300,
                         R=None):
        if R is None:
            R = self.R

        if label is None:
            label = f"{self.A}-{self.B}"

        plt.plot(R, self.fingerprint, label=label)
        plt.axhline(0, 0, 1, color='k', linestyle='--', linewidth=1)
        
        plt.xlabel("Distance (Å)", fontsize=13)
        plt.ylabel('Intensity', fontsize=13)
        
        plt.legend(fontsize=12)

        if save:
            if not os.path.isdir(outdir):
                os.makedirs(outdir, exist_ok=True)
            outfig = os.path.join(
                outdir,f"fingerprint_{self.A}-{self.B}.png")
            plt.savefig(outfig, dpi=dpi)
        if disp:
            plt.show()


class Snapshots:
    def __init__(self, 
                 interval: float,
                 vasprun: str = 'vasprun.xml',
                 prefix: str = 'snapshots'):
        """
        Generate step-wise poscar files
        Args:
            interval (float):
                time interval in ps
            vasprun (str, optional):
                vasprun.xml file. Defaults to 'vasprun.xml'.
            prefix (str, optional):
                name of directory where poscar files will be saved. Defaults to 'snapshots'.
        """
        if os.path.isfile(vasprun):
            self.vasprun = vasprun
        else:
            print(f"Error: {vasprun} is not found.")
            sys.exit(0)
            
        if not os.path.isdir(prefix):
            os.makedirs(prefix)
            print(f'{prefix} directory is created.')
        
        self.interval = interval
        self.prefix = prefix
        
        # read vasprun.xml
        self.nsw = None
        self.pos = None
        self.potim = None
        self.digit = None
        self.lattice = None
        self.nsw_cut = None
        self.num_step = None
        self.num_atom = None
        self.atom_counts = None
        self.interval_nsw = None
        self.read_vasprun()
        
        # atom pairs
        self.pair = []
        self.get_atom_pairs()
        
        # save poscars
        self.save_poscars()
    
    def save_poscars(self):
        for i in range(self.num_step):
            poscar = os.path.join(
                self.prefix, f"POSCAR_{format(i, self.digit)}"
            )
            
            with open(poscar, "w") as f:
                f.write(f"step_{i}. generated by vachoppy.\n")
                f.write("1.0\n")
                for lat in self.lattice:
                    f.write("%.6f %.6f %.6f\n"%(lat[0], lat[1], lat[2]))
                f.write(f"{'  '.join(map(str, self.atom_counts.keys()))}\n")
                f.write(f"{'  '.join(map(str, self.atom_counts.values()))}\n")
                f.write("Direct\n")
                for coord in self.pos[i]:
                    f.write("%.6f %.6f %.6f\n"%(coord[0], coord[1], coord[2]))
                    
    def get_atom_pairs(self):
        atom_species = list(self.atom_counts.keys())
        self.pair.extend(combinations_with_replacement(atom_species, 2))
    
    def read_vasprun(self):
        v = Vasprun(self.vasprun,  
                    parse_dos=False, 
                    parse_eigen=False,
                    parse_potcar_file=False)
        
        self.potim = v.incar.get("POTIM")
        
        structure = v.final_structure
        self.lattice = structure.lattice.matrix.tolist()
        self.num_atom = v.final_structure.num_sites
        atom_symbols = [str(site.specie) for site in structure.sites]
        self.atom_counts = dict(Counter(atom_symbols))
        
        # ionic steps
        iterations = v.ionic_steps
        self.nsw = len(iterations)
        
        # convert interation : ps to iteration
        eps = 1e-9
        val = self.interval * 1000 / self.potim
        if not math.isclose(val, round(val), abs_tol=eps):
            print("ERROR: interval must be a multiple of potim.")
            sys.exit(0)
        self.interval_nsw = round(val)
        
        self.num_step = int(self.nsw / self.interval_nsw)
        self.nsw_cut = self.num_step * self.interval_nsw
        self.digit = f'0{int(np.log10(self.num_step)) + 1}'
        
        # atomic positions
        pos = np.zeros((self.nsw_cut, self.num_atom, 3), dtype=np.float64)
        for step_idx, step in enumerate(iterations[:self.nsw_cut]):
            for i in range(self.num_atom):
                pos[step_idx, i, :] = step["structure"].sites[i].frac_coords
        
        # unwrap position
        displacement = np.zeros_like(pos)
        displacement[1:, :] = np.diff(pos, axis=0)
        displacement[displacement>0.5] -= 1.0
        displacement[displacement<-0.5] += 1.0
        displacement = np.cumsum(displacement, axis=0)
        pos = displacement + pos[0]
        
        # step-averaged coordinatoin
        pos = pos.reshape(self.num_step, self.interval_nsw, self.num_atom, 3)
        pos = np.average(pos, axis=1)
        
        # wrao bacj into cell
        pos = pos - np.floor(pos)
        self.pos = pos

                   
def get_fingerprint(poscar, 
                    filename, 
                    atom_pair, 
                    Rmax, 
                    delta, 
                    sigma):
    # get fingerprint
    fingerprint = []
    for (A, B) in atom_pair:
        fingerprint_i = FingerPrint(A, B, poscar, Rmax, delta, sigma)
        fingerprint.append(fingerprint_i.fingerprint)
    fingerprint = np.array(fingerprint).reshape(1, -1).squeeze()
    
    # save fingerprint
    with open(filename, 'w') as f:
        f.write(f'# Rmax, delta, sigma = {Rmax}, {delta}, {sigma}\n')
        f.write('# pair : ')
        for (A, B) in atom_pair:
            f.write(f'{A}-{B}, ')
        f.write('\n')
        
        R = np.linspace(0, Rmax * len(atom_pair), len(fingerprint))
        for x, y in zip(R, fingerprint):
            f.write(f'  {x:2.6f}\t{y:2.6f}\n')

    return fingerprint


def phase_transition_serial(snapshots,
                            poscar_ref,
                            Rmax,
                            delta,
                            sigma,
                            prefix='fingerprints'):
    
    digit = snapshots.digit
    interval = snapshots.interval
    atom_pair = snapshots.pair
    path_poscar = snapshots.prefix
    task_size = snapshots.num_step
    
    if not os.path.isdir(prefix):
            os.makedirs(prefix)
            print(f'{prefix} directory is created.')
    
    fingerprint_ref = get_fingerprint(
        poscar=poscar_ref,
        filename=os.path.join(prefix, "fingerprint_ref.txt"),
        atom_pair=atom_pair,
        Rmax=Rmax,
        delta=delta,
        sigma=sigma
    )
    
    results = []
    for i in tqdm(list(range(task_size)),
                  bar_format='{l_bar}%s{bar:35}%s{r_bar}{bar:-10b}'%(Fore.GREEN, Fore.RESET),
                  ascii=False,
                  desc=f'{RED}{BOLD}Progress{RESET}'):
        fingerprint = get_fingerprint(
                poscar=os.path.join(path_poscar, f"POSCAR_{format(i, digit)}"),
                filename=os.path.join(prefix, f"fingerprint_{format(i, digit)}.txt"),
                atom_pair=atom_pair,
                Rmax=Rmax,
                delta=delta,
                sigma=sigma
            )
        d_cos = CosineDistance(fingerprint_ref, fingerprint)
        results.append([i, d_cos])
    
    results = np.array(results)
    results[:,0] *= interval
    
    # plot cosine distance
    plt.figure(figsize=(12, 4))
    plt.scatter(results[:,0], results[:,1], s=25)
    plt.xlabel("t (ps)", fontsize=13)
    plt.ylabel('Cosine distnace', fontsize=13)
    plt.savefig('cosine_distance.png', dpi=300)
    plt.close()
    print('cosine_distance.png is created.')
    
    # save cosine distance
    with open('cosine_distance.txt', 'w') as f:
        f.write(f'# Rmax, delta, sigma = {Rmax}, {delta}, {sigma}\n')
        f.write('# pair : ')
        for (A, B) in atom_pair:
            f.write(f'{A}-{B}, ')
        f.write('\n')
        for [x, y] in results:
                f.write(f'  {x}\t{y:.6f}\n')
    print('cosine_distance.txt is created.')
        

def phase_transition_parallel(snapshots,
                              poscar_ref,
                              Rmax,
                              delta,
                              sigma,
                              prefix='fingerprints'):
    """
    snapshots : instant of Snapshots class
    """
    # time estimation
    time_i = time.time()
    
    # parameters from Snapshots
    digit = snapshots.digit
    interval = snapshots.interval
    atom_pair = snapshots.pair
    path_poscar = snapshots.prefix
    task_size = snapshots.num_step
    
    # parallelization
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    
    # master
    if rank == 0:
        if not os.path.isdir(prefix):
            os.makedirs(prefix)
            print(f'{prefix} directory is created.')
            
        task_queue = list(range(task_size))
        print(f"Number of snapshots : {task_size}")
        
        results = []
        completed_task = 0
        terminated_worker = 0
        active_workers = size - 1
        
        while completed_task < task_size or terminated_worker < active_workers:
            status = MPI.Status()
            worker_id, task_result = comm.recv(
                source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status
            )
            
            # tag=4 : response to termination sign
            if status.Get_tag() == 4:
                terminated_worker += 1
                continue
            
            if task_result is not None:
                completed_task += 1
                results.append(task_result)
                print(f"Progress: {completed_task}/{task_size}, " +
                      f"remaining workers = {active_workers - terminated_worker}/{active_workers}")
            
            if task_queue:
                new_task = task_queue.pop()
                # tag=1 : allocating task
                comm.send(new_task, dest=worker_id, tag=1)
            else:
                # tag=0 : termination sign
                comm.send(None, dest=worker_id, tag=0)
        
        # to confirm all workers are terminated
        while terminated_worker < active_workers:
            worker_id, _ = comm.recv(source=MPI.ANY_SOURCE, tag=4)
            terminated_worker += 1
    
    # workers
    else:
        # fingerprint of reference phase
        fingerprint_ref = get_fingerprint(
            poscar=poscar_ref,
            filename=os.path.join(prefix, "fingerprint_ref.txt"),
            atom_pair=atom_pair,
            Rmax=Rmax,
            delta=delta,
            sigma=sigma
        )
            
        # tag=2 : request for very first task
        comm.send((rank, None), dest=0, tag=2)
        
        while True:
            task = comm.recv(source=0, tag=MPI.ANY_TAG)
            
            if task is None:
                # no remaining task
                comm.send((rank, None), dest=0, tag=4)
                break
            
            # get fingerprint
            fingerprint = get_fingerprint(
                poscar=os.path.join(path_poscar, f"POSCAR_{format(task, digit)}"),
                filename=os.path.join(prefix, f"fingerprint_{format(task, digit)}.txt"),
                atom_pair=atom_pair,
                Rmax=Rmax,
                delta=delta,
                sigma=sigma
            )
            
            # get cosine distnace
            d_cos = CosineDistance(fingerprint_ref, fingerprint)
            
            # tag=3 : send result to master
            comm.send((rank, [task, d_cos]), dest=0, tag=3)
            
    if rank == 0:
        # sort results
        results.sort(key=lambda x: x[0])
        results = np.array(results)
        results[:,0] *= interval
        
        # plot cosine distance
        plt.figure(figsize=(12, 4))
        plt.scatter(results[:,0], results[:,1], s=25)
        plt.xlabel("t (ps)", fontsize=13)
        plt.ylabel('Cosine distnace', fontsize=13)
        plt.savefig('cosine_distance.png', dpi=300)
        plt.close()
        print('cosine_distance.png is created.')
        
        # save cosine distance
        with open('cosine_distance.txt', 'w') as f:
            f.write(f'# Rmax, delta, sigma = {Rmax}, {delta}, {sigma}\n')
            f.write('# pair : ')
            for (A, B) in atom_pair:
                f.write(f'{A}-{B}, ')
            f.write('\n')
            for [x, y] in results:
                 f.write(f'  {x}\t{y:.6f}\n')
        print('cosine_distance.txt is created.')
        
        time_f = time.time()
        print(f"total time taken = {time_f - time_i} s")
        
        
        
# class Snapshots:
#     def __init__(self,
#                  xdatcar: str,
#                  outcar: str,
#                  interval: float,
#                  prefix: str ='snapshots'):
#         """
#         Arguements
#         ----------
#         xdatcar : str
#             Path to XDATCAR file
#         outcar : str
#             Path to OUTCAR file
#         interval : float 
#             Time interval (ps)
#         prefix : str, optional
#             Prefix for output files
#         """
#         self._validate_file(xdatcar, "XDATCAR")
#         self._validate_file(outcar, "OUTCAR")

#         self.xdatcar = xdatcar
#         self.outcar = outcar
#         self.prefix = prefix
#         self.interval = interval
        
#         if not os.path.isdir(self.prefix):
#             os.makedirs(self.prefix)
#             print(f'{self.prefix} directory is created.')
            
#         # read outcar
#         self.potim = None
#         self.read_outcar()
        
#         if self.potim is None:
#             print(f"POTIM does not exist in OUTCAR.")
#             sys.exit(0)
            
#         self.interval_nsw = self._compute_interval_nsw(interval, self.potim)
        
#         # read xdatcar
#         self.digit = None
#         self.nsw_cut = None
#         self.num_step = None
#         self.atom_species = None
#         self.lattice_parameter = None
#         self.position = []
#         self.read_xdatcar()
        
#         self.pair = []
#         self.pair.extend(combinations_with_replacement(self.atom_species, 2))
        
#         # save poscar
#         for i in range(self.num_step):
#             filename = os.path.join(
#                 self.prefix, f"POSCAR_{format(i, self.digit)}"
#             )
#             self.save_poscar(i, filename)
            
#         print(f"AIMD snapshots are saved in {self.prefix} directory.")
    
#     def _validate_file(self, path, label="file"):
#         if not os.path.isfile(path):
#             print(f"ERROR: {label} '{path}' is not found.")
#             sys.exit(0)

#     def _compute_interval_nsw(self, interval, potim):
#         eps = 1e-9
#         val = interval * 1000 / potim
#         if math.isclose(val, round(val), abs_tol=eps):
#             return int(round(val))
#         print("ERROR: interval must be a multiple of potim.")
#         sys.exit(0)
                
#     def read_outcar(self):
#         with open(self.outcar, 'r') as f:
#             for line in f:
#                 if 'POTIM' in line:
#                     self.potim = float(line.split()[2])
#                     break
        
#     def read_xdatcar(self):
        
#         with open(self.xdatcar, 'r') as f:
#             lines = np.array([s.strip() for s in f])
        
#         nsw = find_last_direct_line(self.xdatcar)
#         num_step = int(nsw / self.interval_nsw)
#         nsw_cut = num_step * self.interval_nsw
        
#         lattice_parameter = np.array(
#             [s.split() for s in lines[2:5]], dtype=np.float64
#         ) * np.float64(lines[1])
        
#         atom_species = np.array(lines[5].split())
#         num_atoms = np.array(lines[6].split(), dtype=np.int32)
#         num_atoms_tot = np.sum(num_atoms)
        
#         digit = int(np.log10(num_step)) + 1
#         digit = f'0{digit}'
        
#         for i, spec in enumerate(atom_species):           
#             atom = {}
#             atom['species'] = spec
#             atom['num'] = num_atoms[i]
            
#             traj = np.zeros((atom['num'], num_step, 3)) 

#             for j in range(atom['num']):
#                 start = np.sum(num_atoms[:i]) + j + 8
#                 end = 7 + nsw_cut * (num_atoms_tot + 1)
#                 step = num_atoms_tot + 1
#                 coords = [s.split() for s in lines[start:end:step]]
#                 coords = np.array(coords, dtype=float)
                
#                 displacement = np.zeros_like(coords)
#                 displacement[0,:] = 0
#                 displacement[1:,:] = np.diff(coords, axis=0)

#                 # correction for periodic boundary condition
#                 displacement[displacement>0.5] -= 1.0
#                 displacement[displacement<-0.5] += 1.0
#                 displacement = np.cumsum(displacement, axis=0)
#                 coords = coords[0] + displacement

#                 # averaged coordination
#                 coords = coords.reshape(num_step, self.interval_nsw, 3)
#                 coords = np.average(coords, axis=1)

#                 # wrap back into cell
#                 coords = coords - np.floor(coords)
#                 traj[j] = coords
#             atom['traj'] = traj
            
#             self.position += [atom]
        
#         self.digit = digit
#         self.nsw_cut = nsw_cut
#         self.num_step = num_step
#         self.atom_species = atom_species
#         self.lattice_parameter = lattice_parameter
            
#     def save_poscar(self, 
#                     step, 
#                     filename):
        
#         with open(filename, 'w') as f:
#             f.write(f"step_{step}. generated by vachoppy.\n")
#             f.write("1.0\n")

#             for lat in self.lattice_parameter:
#                 f.write("%.6f %.6f %.6f\n"%(lat[0], lat[1], lat[2]))

#             for atom in self.position:
#                 f.write(f"{atom['species']} ")
#             f.write('\n')
            
#             for atom in self.position:
#                 f.write(f"{atom['num']} ")
#             f.write('\n')
            
#             f.write("Direct\n")
            
#             for atom in self.position:
#                 for traj in atom['traj'][:,step,:]:
#                     f.write("%.6f %.6f %.6f\n"%(traj[0], traj[1], traj[2]))



import os
import sys
import copy
import numpy as np
from tqdm import tqdm
from colorama import Fore
import matplotlib.pyplot as plt
from itertools import combinations_with_replacement

from vachoppy.inout import *
from vachoppy.einstein import *
from vachoppy.trajectory import *
from vachoppy.parameter import *
from vachoppy.fingerprint import *
from vachoppy.utils import *

try:
    from mpi4py import MPI
    PARALELL = True
except:
    PARALELL = False

BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m' # Green color
RED = '\033[91m'   # Red color
RESET = '\033[0m'  # Reset to default color


# -m t option
class MakeAnimation:
    def __init__(self,
                 data,
                 temp,
                 label,
                 interval,
                 poscar_lattice,
                 update_alpha=0.75,
                 show_index=False,
                 dpi=300,
                 rmax=3.0,
                 tol=1e-3,
                 tolerance=1e-3,
                 verbose=True):
        
        print("VacHopPy is running...\n")

        if int(temp) in data.temp:
            self.temp = temp
             
        else:
            print(f"{temp}K is not valid.")
            sys.exit(0)
            
        if label in data.label[list(data.temp).index(self.temp)]:
            self.label = label
            
        else:
            print(f"{label} is not valid.")
            sys.exit(0)
            
        for i, temp_i in enumerate(data.temp):
            if abs(self.temp - temp_i) < 1e-9:
                index_temp = i
                break
        
        index_label = data.label[index_temp].index(self.label)
        
        self.pos = data.pos[index_temp][index_label]
        self.cond = data.cond[index_temp][index_label]
        self.force = data.force[index_temp][index_label]
        
        # self.potim = data.potim[index_temp]
        self.interval = interval
        self.poscar_lattice = poscar_lattice
        self.update_alpha = update_alpha
        self.show_index = show_index
        self.dpi = dpi
        self.rmax = rmax
        self.tol = tol
        self.tolerance = tolerance
        self.verbose = verbose
        
        vac_info = VacancyInfo(
            data=data,
            poscar_lattice=poscar_lattice
        )
        
        self.symbol = vac_info.symbol_vac
        self.num_vac = vac_info.number_vac
        print(f"Vacancy type : {self.symbol}")
        print(f"Number of vacancies : {self.num_vac}")
        
        if self.verbose:
            file_out = "trajectory.txt"
            with open(file_out, "w", encoding="utf-8") as f:
                original_stdout = sys.stdout
                sys.stdout = f
                try:
                    self.lattice = Lattice(
                        poscar_lattice=self.poscar_lattice,
                        symbol=self.symbol,
                        rmax=self.rmax,
                        tol=self.tol,
                        tolerance=self.tolerance,
                        verbose=self.verbose
                    )
                finally:
                    sys.stdout = original_stdout
        else:
            self.lattice = Lattice(
                poscar_lattice=self.poscar_lattice,
                symbol=self.symbol,
                rmax=self.rmax,
                tol=self.tol,
                tolerance=self.tolerance,
                verbose=self.verbose
            )
        
        self.traj = Trajectory(
            interval=interval,
            num_vac=self.num_vac,
            lattice=self.lattice,
            pos_file=self.pos,
            force_file=self.force,
            cond_file=self.cond,
            verbose=self.verbose
        )
            
        self.save_animation()
        
        if self.verbose:
            with open(file_out, "a", encoding="utf-8") as f:
                original_stdout = sys.stdout
                sys.stdout = f
                
                print('')
                try:
                    anal = TrajectoryAnalyzer(
                        lattice=self.lattice,
                        trajectory=self.traj,
                        tolerance=self.tolerance,
                        verbose=self.verbose
                    )
                finally:
                    sys.stdout = original_stdout
                    
            print(f"{file_out} is created.")
        
        print('VacHopPy is done.')
            
    def save_animation(self):
        simulation_time = self.traj.nsw * self.traj.potim / 1000
        t_interval = self.traj.interval
        
        print(f'{GREEN}{BOLD}\nInformation on animation{RESET}')
        print(f"    Total simulation time  : {simulation_time:.3f} ps")
        print(f"    Time interval per step : {t_interval:.3f} ps")
        print(f"    Total number of steps  : {self.traj.num_step} (={simulation_time:.3f}/{t_interval:.3f})")
        print('')
        
        print(f'Enter the initial and final steps (min: 0, max: {self.traj.num_step})')
        step = input(f'{MAGENTA}{BOLD}Answer{RESET} (example: 0 500) : ')
    
        try:
            step = list(map(int, step.split())) 
        except:
            print('The step number must be integer.')
            sys.exit(0)
            
        if step[-1] > self.traj.num_step:
            print(f'    The final step should be less than {self.traj.num_step}')
            print(f'    The final step is set to {self.traj.num_step}')
            step[-1] = self.traj.num_step
            
        step = 'all' if step[-1] == -1 else np.arange(step[0], step[-1])
        
        print('\nEnter the fps value for animation')
        fps = input(f'{MAGENTA}{BOLD}Answer{RESET} (example: 10) : ')
        
        try:
            fps = int(fps)
        except:
            print('The fps must be integer.')
            sys.exit(0)
        
        print('')
        self.traj.animation(
            step=step,
            foldername='snapshot',
            update_alpha=self.update_alpha,
            fps=fps,
            dpi=self.dpi,
            label=self.show_index
        )
        
        print('snapshot directory was created.')
    

# -m p option
class EffectiveHoppingParameter:
    def __init__(self, 
                 data, 
                 interval,
                 poscar_lattice, 
                 parallel,
                 file_out='parameter.txt', 
                 rmax=3.0,
                 tol=1e-3,
                 tolerance=1e-3,
                 use_incomplete_encounter=True,
                 inset_correlatoin_factor=True,
                 verbose=True):
        
        self.data = data
        self.interval = interval
        self.poscar_lattice = poscar_lattice
        self.parallel = parallel
        self.file_out = file_out
        self.rmax = rmax
        self.tol = tol
        self.tolerance = tolerance
        self.use_incomplete_encounter = use_incomplete_encounter
        self.inset_correlatoin_factor = inset_correlatoin_factor
        self.verbose = verbose
        
        vac_info = VacancyInfo(
            data=data,
            poscar_lattice=poscar_lattice
        )
        self.symbol = vac_info.symbol_vac
        self.num_vac = vac_info.number_vac
        
        self.lattice = Lattice(
            poscar_lattice=self.poscar_lattice,
            symbol=self.symbol,
            rmax=self.rmax,
            tol=self.tol,
            tolerance=self.tolerance,
            verbose=False
        )
        
        if self.parallel:
            # get effective hopping parameters
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            size = comm.Get_size()
            
            if rank == 0:
                print('VacHopPy is running...')
                print('')
                print(f"Vacancy type : {self.symbol}")
                print(f"Number of vacancies : {self.num_vac}")
                
                f = open('VACHOPPY_PROGRESS', 'w', buffering=1, encoding='utf-8')
                original_stdout = sys.stdout
                sys.stdout = f
                if size < 2:
                    print("number of cpu node shoud be >= 2.")
                    MPI.COMM_WORLD.Abort(1)

            self.results = Automation_parallel(
                data=self.data, 
                lattice=self.lattice, 
                interval=self.interval,
                num_vac=self.num_vac,
                tolerance=self.tolerance,
                use_incomplete_encounter=self.use_incomplete_encounter
            )
                
            if rank == 0:
                sys.stdout = original_stdout
                f.close()
                # merge unknown paths in parallel calculculations
                self.results[-1].lattice = self.merge_lattice()
                self.get_parameters()
                print('VacHopPy is done.')
                
        else:
            print('VacHopPy is running...')
            print('')
            print(f"Vacancy type : {self.symbol}")
            print(f"Number of vacancies : {self.num_vac}")
            
            self.results = Automation_serial(
                data=self.data, 
                lattice=self.lattice, 
                interval=self.interval,
                num_vac=self.num_vac,
                tolerance=self.tolerance,
                use_incomplete_encounter=self.use_incomplete_encounter
            )
            self.get_parameters()
            print('VacHopPy is done.')
            
    def append_unknown_path(self, path, lattice):
        for path_unknown in lattice.path_unknown:
            check1 = True if path['site_init'] == path_unknown['site_init'] else False
            check2 = True if path['site_final'] == path_unknown['site_final'] else False
            check3 = True if abs(path['distance'] - path_unknown['distance']) < self.tolerance else False
            
            if check1 and check2 and check3:
                return path['name'], path_unknown['name']
        
        name_before = path['name']
        path['name'] = f"unknown{len(lattice.path_unknown) + 1}"
        lattice.path_unknown.append(path)
    
        return name_before, path['name']
            
    def merge_lattice(self):
        lattice_merged = None
        for result in self.results:
    
            if lattice_merged is None:
                lattice_merged = copy.deepcopy(result.lattice)
                continue
            
            if int(np.sum(result.counts_unknown)) == 0:
                # no unknown path detected
                continue
            
            name_unknown_before = []
            name_unknown_after = []
            
            for path in result.path_unknown:
                name_before, name_after = self.append_unknown_path(path, lattice_merged)
                
                if name_before == name_after:
                    continue
                
                name_unknown_after.append(name_before)
                name_unknown_after.append(name_after)
            
            # update hopping history
            for history in result.hopping_history:
                for hop in history:
                    if hop['name'] in name_unknown_before:
                        index = name_unknown_before.index(hop['name'])
                        hop['name'] = name_unknown_after[index]
            
            # update unknown_name
            for i, name in enumerate(result.unknown_name):
                if name in name_unknown_before:
                    index = name_unknown_before.index(name)
                    result.unknown_name[i] = name_unknown_after[index]
        
        return lattice_merged
            
    def get_parameters(self):
        with open(self.file_out, 'w', encoding='utf-8') as f:
            original_stdout = sys.stdout
            sys.stdout = f
            
            try:
                extractor=ParameterExtractor(
                    data=self.data,
                    results=self.results,
                    tolerance=self.tolerance,
                    verbose=self.verbose,
                    figure=self.verbose,
                    file_out=None,
                    inset_correlatoin_factor=self.inset_correlatoin_factor
                )
                
            finally:
                sys.stdout = original_stdout  
                
        print('')
        print(f"{self.file_out} is created.")
        print("D_rand.png is created.")
        print("f_cor.png is created.")
        print("tau.png is created.")
        

# -m pp option
class PostEffectiveHoppingParameter:
    def __init__(self,
                 file_params='parameter.txt',
                 file_neb='neb.csv',
                 file_out='postprocess.txt',
                 verbose=True):
        
        self.file_params = file_params
        self.file_neb = file_neb
        self.file_out = file_out
        self.verbose = verbose
        
        with open(self.file_out, 'w', encoding='utf-8') as f:
            original_stdout = sys.stdout
            sys.stdout = f
            
            try:
                postprocess = PostProcess(
                    file_params=self.file_params,
                    file_neb=self.file_neb,
                    verbose=self.verbose
                )
                
            finally:
                sys.stdout = original_stdout
                
        print(f"{self.file_out} is created.")


# -m f option
class PhaseTransition:
    def __init__(self,
                 vasprun,
                 interval,
                 Rmax,
                 delta,
                 sigma,
                 parallel,
                 poscar_ref='POSCAR_REF',
                 prefix1='snapshots',
                 prefix2='fingerprints'):
        
        self.vasprun = vasprun
        self.interval = interval
        self.Rmax = Rmax
        self.delta = delta
        self.sigma = sigma
        self.parallel = parallel
        self.poscar_ref = poscar_ref
        self.prefix1 = prefix1
        self.prefix2 = prefix2
        
        if self.parallel:
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            size = comm.Get_size()
            
            if rank == 0:
                print('VacHopPy is running...\n')
                f = open('VACHOPPY_PROGRESS', 'w', buffering=1, encoding='utf-8')
                original_stdout = sys.stdout
                sys.stdout = f
                if size < 2:
                    print("number of cpu node shoud be >= 2.")
                    MPI.COMM_WORLD.Abort(1)
                try:
                    snapshots = Snapshots(
                        interval=self.interval,
                        vasprun=self.vasprun,
                        prefix=self.prefix1
                    )
                    
                except SystemExit:
                    print("Error occured duing instantiating fingerprint.Snapshots.")
                    MPI.COMM_WORLD.Abort(1)
                    
                except Exception as e:
                    print(f"Unexpected error: {e}")
                    MPI.COMM_WORLD.Abort(1)
                
            else:
                snapshots = None
            
            snapshots = comm.bcast(snapshots, root=0)
            
            phase_transition_parallel(
                snapshots=snapshots,
                poscar_ref=self.poscar_ref,
                Rmax=self.Rmax,
                delta=self.delta,
                sigma=self.sigma,
                prefix=self.prefix2
            )
            
            if rank==0:
                sys.stdout = original_stdout
                f.close()
                print('VacHopPy is done.')
        
        else:
            print('VacHopPy is running...\n')
            snapshots = Snapshots(
                interval=self.interval,
                vasprun=self.vasprun,
                prefix=self.prefix1
            )
            phase_transition_serial(
                snapshots=snapshots,
                poscar_ref=self.poscar_ref,
                Rmax=self.Rmax,
                delta=self.delta,
                sigma=self.sigma,
                prefix=self.prefix2
            )
            print('VacHopPy is done.')
            
            
# -u fingerprint option
class GetFingerPrint:
    def __init__(self, 
                 poscar,
                 Rmax,
                 delta,
                 sigma,
                 prefix='fingerprint', 
                 disp=False):
        
        self.poscar = poscar
        self.Rmax = Rmax
        self.delta = delta
        self.sigma = sigma
        self.prefix = prefix
        self.disp = disp
        
        if not os.path.isfile(self.poscar):
            print(f'{self.poscar} is not found')
            sys.exit(0)
        
        # read poscar
        self.atom = None
        self.read_poscar()
        
        self.pair = []
        self.pair.extend(combinations_with_replacement(self.atom, 2))
        
        # fingerprint
        self.fingerprint = []
        self.get_fingerprint()
        self.fingerprint = np.array(self.fingerprint)
        
        # concat fingerprints
        self.fingerprint_concat = self.fingerprint.reshape(1,-1).squeeze()
        
        # save fingerprint
        self.save_fingerprint()
        
        
    def read_poscar(self):
        with open(self.poscar, 'r') as f:
            lines = [line for line in f]
        self.atom = lines[5].split()
    
    def get_pair(self):
        pair = input('input A and B (ex. Hf-O / Hf-Hf,Hf-O / all) : ')
        pair = pair.replace(" ", "")
        
        if pair == 'all':
            self.pair.extend(combinations_with_replacement(self.atom, 2))
            
        else:
            pair = pair.split(',')
            
            for p in pair:
                atoms = p.split('-')
                
                if len(atoms) == 2 and all(atom in self.atom for atom in atoms):
                    self.pair.append(tuple(atoms))
                    
                else:
                    print(f'Invalid pair : {p}')
                    sys.exit(0)
                    
    def get_params(self):
        params = input("input Rmax, delta, and sigma (ex. 15, 0.01, 0.3) : ")
        params = list(map(float, params.replace(" ", "").split(',')))
        self.Rmax, self.delta, self.sigma = params[0], params[1], params[2]
        
    def get_fingerprint(self):
        for (A, B) in self.pair:
            finger = FingerPrint(A, B, 
                                 self.poscar,
                                 self.Rmax, self.delta, self.sigma)
            self.fingerprint.append(finger.fingerprint)
    
    def save_fingerprint(self):
        # save figure
        R = np.linspace(0, self.Rmax, len(self.fingerprint[0]))
        for i in range(len(self.pair)):
            x = R + i*self.Rmax*np.ones_like(R)
            plt.plot(x, self.fingerprint[i], label=f'{self.pair[i][0]}-{self.pair[i][1]}')
        
        plt.axhline(0, 0, 1, color='k', linestyle='--', linewidth=1)
        plt.xlabel("r (Å)", fontsize=13)    
        plt.ylabel('Intensity', fontsize=13)
        plt.legend(fontsize=12)
        plt.savefig(f'{self.prefix}.png', dpi=300)
        print(f'{self.prefix}.png is created.')
        if self.disp:
            plt.show()
        plt.close()
        
        R = np.linspace(0, self.Rmax * len(self.pair), len(self.fingerprint_concat))
        with open(f'{self.prefix}.txt', 'w') as f:
            f.write(f'# Rmax, delta, sigma = {self.Rmax}, {self.delta}, {self.sigma}\n')
            f.write('# pair : ')
            for (A, B) in self.pair:
                f.write(f'{A}-{B}, ')
            f.write('\n')
            for x, y in zip(R, self.fingerprint_concat):
                f.write(f'  {x:2.6f}\t{y:2.6f}\n')
        print(f'{self.prefix}.txt is created.')


# -u cosine_distance option
class GetCosineDistance:
    def __init__(self, fp1, fp2):
        if not os.path.isfile(fp1):
            print(f'{fp1} is not found')
            sys.exit(0)
        if not os.path.isfile(fp2):
            print(f'{fp2} is not found')
            sys.exit(0)
        
        self.fp1 = np.loadtxt(fp1, skiprows=2)[:,1]
        self.fp2 = np.loadtxt(fp2, skiprows=2)[:,1]
        
        if self.fp1.shape != self.fp2.shape:
            print('Size of two fingerprints should be the same')
            sys.exit(0)
        
        self.d_cos = CosineDistance(self.fp1, self.fp2)
        print(f'd_cos = {self.d_cos}')
   
            
# -m e option
class MSD:
    def __init__(self,
                 data,
                 tmax,
                 skip=0,
                 start=1,
                 symbol='O',
                 x_vac=1):
        '''
        Calculate MSD at temperatures in data.
        Arguments:
            data : instance of dataInfo
            tmax : x-axis of msd plot (ps)
            skip : steps to be skipped (ps)
            start :  (ps)
        '''
        print(f'{CYAN}{BOLD}Diffusion coefficient from Einstein relation.{RESET}')
        self.data = data
        self.tmax = tmax
        self.skip = skip
        self.start = start
        self.symbol = symbol
        self.x_vac = x_vac
        
        self.msd = []
        self.D = []
        self.runEnsembleEinstein()
        self.saveMSD()
        self.D /= x_vac
        
        self.plot_results()
        
    def runEnsembleEinstein(self):
        desc = 'Einstein'
        for i, T in enumerate(
            tqdm(self.data.temp, 
                 bar_format='{l_bar}%s{bar:35}%s{r_bar}{bar:-10b}'% (Fore.GREEN, Fore.GREEN),
                 ascii=False,
                 desc=f'{GREEN}{BOLD}{desc:>9s}')):
            path_dir = os.path.join(self.data.prefix1, f'{self.data.prefix2}.{T}K')
            step_skip = int(self.skip * 1000 / self.data.potim[i])
            step_tmax = int(self.tmax * 1000 / self.data.potim[i])
            step_start = int(self.start * 1000 / self.data.potim[i])
            step_end = int(self.tmax * 1000 / self.data.potim[i])
            
            if (self.data.nsw[i] - step_skip) % step_tmax != 0:
                print(f'The MD time is not divided by the tmax. (T={T} K)')
                time_tot = self.data.nsw[i] * self.data.potim[i] / 1000
                print(f'  total MD time = {time_tot} ps ({self.data.nsw[i]} step)')
                print(f'  skip = {self.skip} ps ({step_skip} step)')
                print(f'  tmax = {self.tmax} ps ({step_tmax} step)')
                print(f'  ({self.data.nsw[i]} - {step_skip}) % {step_tmax} != 0')
                sys.exit(0)
            else:
                segment = int((self.data.nsw[i] - step_skip) / step_tmax)
                
            msd = EnsembleEinstein(
                symbol=self.symbol,
                prefix=path_dir,
                labels=self.data.label[i],
                segments=segment,
                skip=step_skip,
                start=step_start,
                end=step_end
            )
            self.msd.append(msd)
    
    def saveMSD(self):
        for msd in self.msd:
            msd.save_msd()
            self.D.append(msd.diffcoeff)
        self.D = np.array(self.D)
        print('msd directory is created.')
        print('msd raw data is written in msd directory.')

    def plot_results(self):
        kb = 8.61733326e-5
        cmap = plt.get_cmap("Set1")
        
        # msd plot
        plt.rcParams['figure.figsize'] = (3.8, 3.8)
        plt.rcParams['font.size'] = 11
        fig, ax = plt.subplots()
        for axis in ['top','bottom','left','right']:
            ax.spines[axis].set_linewidth(1.2)
            
        for i, msd in enumerate(self.msd):
            x, y = msd.timestep, msd.msd
            x_fit = np.linspace(self.start, self.tmax, 1000)
            slop, intercept = np.polyfit(x, y, deg=1)
            ax.plot(x, y, 
                    c=cmap(i), linestyle='-', linewidth=2.5, alpha=0.35)
            ax.plot(x_fit, slop*x_fit+intercept, 
                    c=cmap(i), linestyle='-', linewidth=2.5, label=str(self.data.temp[i]))
        
        ax.axvline(self.start, 0, 1, c='k', linewidth=1, linestyle=':')
        plt.xlim([-1, self.tmax])
        plt.xlabel('t (ps)', fontsize=14)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=14)
        plt.legend(loc='upper left', fontsize=12,
                   title='T (K)', title_fontsize=13, 
                   fancybox=True, framealpha=1, edgecolor='inherit')
        plt.savefig('msd.svg', transparent=True, dpi=600, bbox_inches="tight")
        # plt.show()
        print('msd.svg is created.')
            
        # Arrhenius plot
        plt.rcParams['figure.figsize'] = (3.8, 3.8)
        plt.rcParams['font.size'] = 11
        fig, ax = plt.subplots()
        for axis in ['top','bottom','left','right']:
            ax.spines[axis].set_linewidth(1.2)
            
        slop, intercept = np.polyfit(1/self.data.temp, np.log(self.D), deg=1)
        x_fit = np.linspace(1/self.data.temp[-1], 1/self.data.temp[0], 100)
        plt.plot(x_fit, slop*x_fit+intercept, 'k:')
        
        for i, D in enumerate(self.D):
            plt.scatter(1/self.data.temp[i], np.log(D), 
                        color=cmap(i), marker='s', s=50, label=str(self.data.temp[i]))
        plt.xlabel('1/T (1/K)', fontsize=14)
        plt.ylabel(r'ln $D$', fontsize=14)
        num_data = len(self.D)
        ncol = int(np.ceil(num_data / 5))
        plt.legend(loc='best', fancybox=True, framealpha=1, edgecolor='inherit',
                   ncol=ncol, labelspacing = 0.3, columnspacing=0.5, borderpad=0.2, handlelength=0.6,
                   fontsize=11, title='T (K)', title_fontsize=11)
        
        if num_data >= 3:
            x = np.array([self.data.temp[0], self.data.temp[int(num_data/2)] ,self.data.temp[-1]])
        else:
            x = self.data.temp
            
        x_str = [f'1/{T}' for T in x]
        x = 1/x
        plt.xticks(x, x_str)
        plt.savefig('Arrhenius.svg', transparent=True, dpi=600, bbox_inches="tight")
        # plt.show()
        print('Arrhenius.svg is created.')
        
        # write results
        with open('Einstein.txt', 'w') as f:
            f.write(f'symbol of moving atom = {self.symbol}\n')
            f.write(f'x_vac = {self.x_vac :.5f}\n')
            if self.x_vac == 1:
                f.write('target : atom\n')
            else:
                f.write('target : vacancy\n')
            f.write('\nParameters for diffusion coefficient : \n')
            f.write(f'  D0   = {np.exp(intercept):.5e} m2/s\n')
            f.write(f'  Ea_D = {-kb * slop:.5f} eV\n')
            f.write('\n')
            f.write('Raw data\n')
            f.write('T(K) \tD(m2/s)\n')
            for temp, D in zip(self.data.temp, self.D):
                f.write(f'{temp} \t{D :.5e}\n')
                   
        print('Einstein.txt is created.')
        print('parameters for diffusion coefficient : ')
        print(f'  D0   = {np.exp(intercept) :.5e} m2/s')
        print(f'  Ea_D = {-kb * slop :.5f} eV')
        print('')